import 'dart:convert';

import 'package:contentsquare/contentsquare.dart';
import 'package:contentsquare_platform_interface/contentsquare_platform_interface.dart';
import 'package:flutter/services.dart';
import 'package:flutter_test/flutter_test.dart';

void main() {
  setUpAll(() {
    TestWidgetsFlutterBinding.ensureInitialized();
  });

  group('$Contentsquare', () {
    late MethodChannel channel;
    late List<MethodCall> log;
    late Contentsquare contentsquare;

    setUp(() {
      channel = const MethodChannel('contentsquare');
      log = <MethodCall>[];
      contentsquare = Contentsquare();
      Contentsquare.channel = channel;

      channel.setMockMethodCallHandler((MethodCall methodCall) async {
        log.add(methodCall);
        if (methodCall.method == 'getUserId') {
          return 'Test User Id';
        }
      });
    });

    test('send', () async {
      await contentsquare.send('Test screen');
      expect(log, <Matcher>[
        isMethodCall('send', arguments: {'screenName': 'Test screen'}),
      ]);
    });

    test('sendTransaction', () async {
      await contentsquare.sendTransaction(42, 'EUR', 'testId');
      expect(log, <Matcher>[
        isMethodCall('sendTransaction', arguments: {
          'price': 42,
          'currency': 'EUR',
          'transactionId': 'testId'
        }),
      ]);
    });

    test('sendDynamicVar', () async {
      await contentsquare.sendDynamicVar('key', intValue: 42);
      await contentsquare.sendDynamicVar('key', stringValue: 'test');

      expect(log, <Matcher>[
        isMethodCall('sendDynamicVar', arguments: {
          'key': 'key',
          'stringValue': null,
          'intValue': 42,
        }),
        isMethodCall('sendDynamicVar', arguments: {
          'key': 'key',
          'stringValue': 'test',
          'intValue': null,
        }),
      ]);
    });

    test('optIn', () async {
      await contentsquare.optIn();

      expect(log, <Matcher>[
        isMethodCall('optIn', arguments: null),
      ]);
    });

    test('optOut', () async {
      await contentsquare.optOut();

      expect(log, <Matcher>[
        isMethodCall('optOut', arguments: null),
      ]);
    });

    test('forgetMe', () async {
      await contentsquare.forgetMe();

      expect(log, <Matcher>[
        isMethodCall('forgetMe', arguments: null),
      ]);
    });

    test('getUserId', () async {
      final userId = await contentsquare.getUserId();

      expect(log, <Matcher>[
        isMethodCall('getUserId', arguments: null),
      ]);

      expect(userId, 'Test User Id');
    });

    test('stopTracking', () async {
      await contentsquare.stopTracking();

      expect(log, <Matcher>[
        isMethodCall('stopTracking', arguments: null),
      ]);
    });

    test('resumeTracking', () async {
      await contentsquare.resumeTracking();

      expect(log, <Matcher>[
        isMethodCall('resumeTracking', arguments: null),
      ]);
    });

    test('sendTouchEvent', () async {
      await contentsquare.sendTouchEvent(
        version: 1,
        type: ContentsquareGesture.longPress,
        data: const TouchEventData(path: 'path/to/my/widget'),
      );

      expect(log, <Matcher>[
        isMethodCall('sendTouchEvent', arguments: {
          'body': jsonEncode({
            'type': 8,
            'version': 1,
            'data': {
              'path': 'path/to/my/widget',
              'distance': null,
              'direction': null,
              'velocity': null
            },
          }),
        }),
      ]);
    });
  });
}
