import 'package:contentsquare/src/modules/screengraph/element_extension.dart';
import 'package:contentsquare/src/modules/screengraph/widgets_binding.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';
import 'package:flutter_test/flutter_test.dart';

main() {
  group(
    'ElementExtension.directRenderObject',
    () {
      testWidgets(
        'Should return null when the element is not a RenderObjectElement',
        (WidgetTester tester) async {
          const _key = Key('text');

          await tester.pumpWidget(
            const MaterialApp(
              home: Text(
                'Hello',
                key: _key,
              ),
            ),
          );

          final element = tester.element(find.byKey(_key));

          expect(element.directRenderObject, null);
        },
      );

      testWidgets(
        'Should return the RenderObject associated to the given element when the element is a RenderObjectElement',
        (WidgetTester tester) async {
          const _key = Key('text');

          await tester.pumpWidget(
            const MaterialApp(
              home: ClipRRect(key: _key),
            ),
          );

          final element = tester.element(find.byKey(_key));

          expect(element.directRenderObject, isA<RenderClipRRect>());
        },
      );
    },
  );
  group(
    'ElementExtension.containsGlobalPosition',
    () {
      testWidgets(
        'Should return [MaybeResult.yes] when the element global frame contains the global position',
        (WidgetTester tester) async {
          const _topHalfKey = ValueKey('TopHalf');
          await tester.pumpWidget(
            MaterialApp(
              home: Column(
                children: const <Widget>[
                  Expanded(
                    child: SizedBox.expand(key: _topHalfKey),
                  ),
                  Expanded(child: SizedBox.expand()),
                ],
              ),
            ),
          );

          final screenSize =
              WidgetsBindingExtension.safeInstance!.renderView.size;
          final topQuarterCenter = Offset(
            screenSize.width / 2,
            screenSize.height / 4,
          );

          final topHalfElement = tester.element(find.byKey(_topHalfKey));

          expect(
            topHalfElement.containsGlobalPosition(topQuarterCenter),
            MaybeResult.yes,
          );
        },
      );

      testWidgets(
        'Should return [MaybeResult.no] when the element global frame does not contain the global position',
        (WidgetTester tester) async {
          const _topHalfKey = ValueKey('TopHalf');
          await tester.pumpWidget(
            MaterialApp(
              home: Column(
                children: const <Widget>[
                  Expanded(
                    child: SizedBox.expand(key: _topHalfKey),
                  ),
                  Expanded(child: SizedBox.expand()),
                ],
              ),
            ),
          );

          final screenSize =
              WidgetsBindingExtension.safeInstance!.renderView.size;
          final bottomQuarterCenter = Offset(
            screenSize.width / 2,
            3 * screenSize.height / 4,
          );

          final topHalfElement = tester.element(find.byKey(_topHalfKey));

          expect(
            topHalfElement.containsGlobalPosition(bottomQuarterCenter),
            MaybeResult.no,
          );
        },
      );

      testWidgets(
        'Should return [MaybeResult.unknown] when the element does not have a direct render object',
        (WidgetTester tester) async {
          const _screenKey = ValueKey('Screen');
          await tester.pumpWidget(
            MaterialApp(
              home: Container(
                key: _screenKey,
                child: const Center(
                  child: Text('Hello'),
                ),
              ),
            ),
          );

          final screenSize =
              WidgetsBindingExtension.safeInstance!.renderView.size;
          final center = Offset(screenSize.width / 2, screenSize.height / 2);

          final screenElement = tester.element(find.byKey(_screenKey));

          expect(
            screenElement.containsGlobalPosition(center),
            MaybeResult.unknown,
          );
        },
      );
    },
  );
}
