import 'package:contentsquare/src/modules/screengraph/element_tree.dart';
import 'package:contentsquare/src/modules/screengraph/element_visitor.dart';
import 'package:flutter/material.dart';
import 'package:flutter_test/flutter_test.dart';

class _OverlayEntryWidget extends StatelessWidget {
  const _OverlayEntryWidget({Key? key, required this.child}) : super(key: key);

  final Widget child;

  @override
  Widget build(BuildContext context) {
    return child;
  }
}

main() {
  group(
    'ElementVisitor.root',
    () {
      // We must create a [testWidgets] because [getRootElement] depends on the
      // fact that the widget tree exists and has been built.
      testWidgets(
        'Should return the root element visitor with the excpected attributes',
        (WidgetTester tester) async {
          await tester.pumpWidget(const SizedBox.shrink());

          final rootElementVisitor = CSElementVisitor.root();

          final expectedRootElementVisitor = CSElementVisitor.createForTesting(
            element: getRootElement(),
            tvp: getRootElementPath(),
            childrenMayBeOverlayEntryWidgetOfNavigator: false,
          );

          expect(rootElementVisitor.equals(expectedRootElementVisitor), true);
        },
      );
    },
  );

  group(
    'visitChildren',
    () {
      testWidgets(
        'Should call visit on each child element when visit always returns false',
        (WidgetTester tester) async {
          const _columnKey = ValueKey('Column');
          const _childrenKeys = [
            ValueKey('Child1'),
            ValueKey('Child2'),
            ValueKey('Child3'),
          ];
          await tester.pumpWidget(
            MaterialApp(
              home: Column(
                key: _columnKey,
                children: <Widget>[
                  SizedBox(key: _childrenKeys[0]),
                  SizedBox(key: _childrenKeys[1]),
                  SizedBox(key: _childrenKeys[2]),
                ],
              ),
            ),
          );

          final columnElement = tester.element(find.byKey(_columnKey));
          final columnElementVisitor = CSElementVisitor.createForTesting(
            element: columnElement,
            tvp: 'ColumnTVP',
            childrenMayBeOverlayEntryWidgetOfNavigator: false,
          );

          var visitedChildrenKeys = <ValueKey>[];
          columnElementVisitor.visitChildren(
            visit: (childElementVisitor) {
              visitedChildrenKeys.add(
                childElementVisitor.element.widget.key as ValueKey,
              );

              return false;
            },
          );

          expect(visitedChildrenKeys, _childrenKeys);
        },
      );

      testWidgets(
        'Should stop calling visit on child element when visit returns true',
        (WidgetTester tester) async {
          const _columnKey = ValueKey('Column');
          await tester.pumpWidget(
            MaterialApp(
              home: Column(
                key: _columnKey,
                children: const <Widget>[
                  SizedBox(),
                  SizedBox(),
                  SizedBox(),
                ],
              ),
            ),
          );

          final columnElement = tester.element(find.byKey(_columnKey));
          final columnChildrenElements = <Element>[];
          columnElement.visitChildren(columnChildrenElements.add);

          final columnElementVisitor = CSElementVisitor.createForTesting(
            element: columnElement,
            tvp: 'ColumnTVP',
            childrenMayBeOverlayEntryWidgetOfNavigator: false,
          );
          var visitedChildren = <Element>[];
          columnElementVisitor.visitChildren(
            visit: (childElementVisitor) {
              visitedChildren.add(childElementVisitor.element);

              // Return true to stop visiting more children.
              return true;
            },
          );

          expect(visitedChildren.length, 1);
          expect(visitedChildren.first, columnChildrenElements.first);
        },
      );

      testWidgets(
        'Should create a new elementTVP for each child element',
        (WidgetTester tester) async {
          const _columnKey = ValueKey('Column');
          await tester.pumpWidget(
            MaterialApp(
              home: Column(
                key: _columnKey,
                children: const <Widget>[
                  SizedBox(),
                  SizedBox(),
                  ColoredBox(color: Colors.red),
                ],
              ),
            ),
          );

          final columnElement = tester.element(find.byKey(_columnKey));
          final columnElementVisitor = CSElementVisitor.createForTesting(
            element: columnElement,
            tvp: 'ColumnTVP',
            childrenMayBeOverlayEntryWidgetOfNavigator: false,
          );

          var visitedChildrenTVP = <String>[];
          columnElementVisitor.visitChildren(
            visit: (childElementVisitor) {
              visitedChildrenTVP.add(childElementVisitor.tvp);

              return false;
            },
          );

          expect(
            visitedChildrenTVP,
            [
              'ColumnTVP>SizedBox:eq(0)',
              'ColumnTVP>SizedBox:eq(1)',
              'ColumnTVP>ColoredBox:eq(0)',
            ],
          );
        },
      );

      group(
        'RenderSliverMultiBoxAdaptor',
        () {
          testWidgets(
            'Should use the index of [RenderIndexedSemantics] children',
            (WidgetTester tester) async {
              const itemHeight = 200.0;
              await tester.pumpWidget(
                MaterialApp(
                  home: ListView(
                    children: List.generate(
                      400,
                      (index) => SizedBox(
                        key: ValueKey('SizedBox_$index'),
                        height: itemHeight,
                      ),
                    ),
                  ),
                ),
              );

              // Scroll so that the default index does not match the index of the
              // [RenderIndexedSemantics]s.
              await tester.scrollUntilVisible(
                find.byKey(const ValueKey('SizedBox_30')),
                itemHeight,
              );

              final columnElement = tester.element(
                find.byElementType(SliverMultiBoxAdaptorElement),
              );
              final columnElementVisitor = CSElementVisitor.createForTesting(
                element: columnElement,
                tvp: 'ListViewTVP',
                childrenMayBeOverlayEntryWidgetOfNavigator: false,
              );

              var visitedChildrenTVP = <String>[];
              columnElementVisitor.visitChildren(
                visit: (columnElementVisitor) {
                  visitedChildrenTVP.add(columnElementVisitor.tvp);

                  return false;
                },
              );

              expect(
                visitedChildrenTVP,
                [
                  'ListViewTVP>KeyedSubtree:eq(26)',
                  'ListViewTVP>KeyedSubtree:eq(27)',
                  'ListViewTVP>KeyedSubtree:eq(28)',
                  'ListViewTVP>KeyedSubtree:eq(29)',
                  'ListViewTVP>KeyedSubtree:eq(30)',
                  'ListViewTVP>KeyedSubtree:eq(31)',
                  'ListViewTVP>KeyedSubtree:eq(32)',
                ],
              );
            },
          );
        },
      );

      group(
        '_OverlayEntryWidget',
        () {
          testWidgets(
            'Should remove the index from [_OverlayEntryWidget] when its parent is a [Navigator]',
            (WidgetTester tester) async {
              await tester.pumpWidget(
                Directionality(
                  textDirection: TextDirection.ltr,
                  child: Navigator(
                    pages: const [
                      // Use 2 pages to ensure that we don't remove eq because its
                      // the first element
                      MaterialPage(child: Text('Page1')),
                      MaterialPage(child: Text('Page2')),
                    ],
                    onPopPage: (route, result) => route.didPop(result),
                  ),
                ),
              );

              final theatreElement = tester.firstElement(
                find.byElementPredicate(
                  (element) =>
                      element.widget.runtimeType.toString() == '_Theatre',
                ),
              );
              final theatreElementVisitor = CSElementVisitor.createForTesting(
                element: theatreElement,
                tvp: 'TheatreTVP',
                childrenMayBeOverlayEntryWidgetOfNavigator: true,
              );

              var visitedChildrenTVP = <String>[];
              theatreElementVisitor.visitChildren(
                visit: (childElementVisitor) {
                  visitedChildrenTVP.add(childElementVisitor.tvp);

                  return false;
                },
              );

              // Each [MaterialPage] creates a [Route] which extends [ModalRoute]
              // Each ModalRoute creates 2 [OverlayEntry]s: The first is marked
              // opaque while the second has maintainState set to true
              // The Overlay algorithm, which gives its children to _Theater,
              // works by taking the reverse list of its overlay entries and do
              // has follow:
              //   - Add every children until an opaque children is found
              //   - For each following child, add it if and only if it has
              //     maintainState set to true
              //
              // The first [_OverlayEntryWidget] corresponds to the 1st
              // [OverlayEntry] (of the 1st page) which is opaque
              // The second [_OverlayEntryWidget] corresponds to the 2nd
              // [OverlayEntry] (of the 1st page) which has maintainState set to true
              // The third [_OverlayEntryWidget] corresponds to the 2nd
              // [OverlayEntry] (of the 2nd page) which has maintainState set to true
              // Note: The 1st [OverlayEntry] (of the 2nd page) is below an opaque
              // entry and does NOT have maintainState set to true, therefore it is not
              // included
              expect(
                visitedChildrenTVP,
                [
                  'TheatreTVP>_OverlayEntryWidget',
                  'TheatreTVP>_OverlayEntryWidget',
                  'TheatreTVP>_OverlayEntryWidget',
                ],
              );
            },
          );

          testWidgets(
            'Should not remove the index from [_OverlayEntryWidget] when it is a widget created by something other than the Navigator, even if a Navigator is in its parent widget tree',
            (WidgetTester tester) async {
              const columnKey = ValueKey('Column');
              await tester.pumpWidget(
                Directionality(
                  textDirection: TextDirection.ltr,
                  child: Navigator(
                    pages: [
                      MaterialPage(
                        child: Column(
                          key: columnKey,
                          children: const [
                            // Use 2 _OverlayEntryWidget to ensure that we don't
                            // simply default to eq(0)
                            _OverlayEntryWidget(child: Text('0')),
                            _OverlayEntryWidget(child: Text('1')),
                          ],
                        ),
                      ),
                    ],
                    onPopPage: (route, result) => route.didPop(result),
                  ),
                ),
              );

              final columnElement = tester.firstElement(find.byKey(columnKey));
              final columnElementVisitor = CSElementVisitor.createForTesting(
                element: columnElement,
                tvp: 'ColumnTVP',
                childrenMayBeOverlayEntryWidgetOfNavigator: false,
              );

              var visitedChildrenTVP = <String>[];
              columnElementVisitor.visitChildren(
                visit: (childElementVisitor) {
                  visitedChildrenTVP.add(childElementVisitor.tvp);

                  return false;
                },
              );

              expect(
                visitedChildrenTVP,
                [
                  'ColumnTVP>_OverlayEntryWidget:eq(0)',
                  'ColumnTVP>_OverlayEntryWidget:eq(1)',
                ],
              );
            },
          );
        },
      );
    },
  );
}
