import 'dart:math';

import 'package:contentsquare/src/modules/screengraph/hit_test/get_tvp_at_hit_position.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';
import 'package:flutter_test/flutter_test.dart';

import '../../../helpers/decompress_tvp.dart';
import '../../../helpers/matchers/fixtures/matcher.dart';
import '../../../helpers/get_center.dart';

class OutOfBoundsHit extends SingleChildRenderObjectWidget {
  const OutOfBoundsHit({Key? key, required this.size}) : super(key: key);

  final Size size;

  @override
  RenderObject createRenderObject(BuildContext context) =>
      RenderOutOfBoundsHit(customSize: size);
}

class RenderOutOfBoundsHit extends RenderBox {
  RenderOutOfBoundsHit({required this.customSize});

  final Size customSize;

  @override
  void performLayout() {
    size = Size(
      min(customSize.width, constraints.maxWidth),
      min(customSize.height, constraints.maxHeight),
    );
  }

  @override
  bool hitTest(BoxHitTestResult result, {required Offset position}) {
    result.add(BoxHitTestEntry(this, position));
    return true;
  }
}

void main() {
  group(
    'getTVPAtHitPosition',
    () {
      testWidgets(
        'Should return the TVP of the last hittable widget around this position and ignore the inner Text which is not hittable',
        (WidgetTester tester) async {
          const _key = ValueKey('Text');
          await tester.pumpWidget(
            MaterialApp(
              home: Scaffold(
                // [GestureDetector] will capture the tap but the TVP should extends to
                // the [SizedBox].
                body: GestureDetector(
                  child: const Center(
                    child: SizedBox(
                      width: 100,
                      height: 100,
                      // Text is not hittable.
                      child: Text(
                        'Text',
                        key: _key,
                      ),
                    ),
                  ),
                ),
              ),
            ),
          );

          final hitTVPs = getTVPAtHitPosition(
            getCenterFromRenderBoxKey(key: _key),
          );
          final decompressedHitTVPs = hitTVPs
              .map(
                (hitTVP) => decompressTVP(tvp: hitTVP),
              )
              .toList();

          expect(
            decompressedHitTVPs.single,
            matchesFixture('hitTVPNotHittable'),
          );
        },
      );

      group(
        'Scrollables',
        () {
          testWidgets(
            'Should return the right TVP when the widget tapped is in a scrollable',
            (WidgetTester tester) async {
              await tester.pumpWidget(
                MaterialApp(
                  home: ListView.builder(
                    itemBuilder: (_, index) => SizedBox(
                      height: 50,
                      child: Text(
                        'Text $index',
                        key: ValueKey('Text $index'),
                      ),
                    ),
                    itemCount: 2000,
                  ),
                ),
              );

              final hitTVPs = getTVPAtHitPosition(
                getCenterFromRenderBoxKey(key: const ValueKey('Text 3')),
              );
              final decompressedHitTVPs =
                  hitTVPs.map((hitTVP) => decompressTVP(tvp: hitTVP)).toList();

              expect(
                decompressedHitTVPs.single,
                matchesFixture('hitTVPWithinScrollable'),
              );
            },
          );

          testWidgets(
            'Should return the right TVP when the widget tapped is in a scrollable and the scrollable needs to be scrolled to reveal the widget',
            (WidgetTester tester) async {
              await tester.pumpWidget(
                MaterialApp(
                  home: ListView.builder(
                    itemBuilder: (context, index) => SizedBox(
                      height: MediaQuery.of(context).size.height / 10,
                      child: Text(
                        'Text $index',
                        key: ValueKey('Text $index'),
                      ),
                    ),
                    itemCount: 30,
                  ),
                ),
              );

              const _itemKey = ValueKey('Text 20');
              final itemFinder = find.byKey(_itemKey);
              await tester.scrollUntilVisible(itemFinder, 100);

              final hitTVPs = getTVPAtHitPosition(
                getCenterFromRenderBoxKey(key: _itemKey),
              );
              final decompressedHitTVPs =
                  hitTVPs.map((hitTVP) => decompressTVP(tvp: hitTVP)).toList();

              expect(
                decompressedHitTVPs.single,
                matchesFixture('hitTVPNeedsScroll'),
              );
            },
          );
        },
      );

      testWidgets(
        'Should return one TVP of the parent which is in the screengraph when several widgets have the same parent in the screengraph and are not themselves in the screengraph\n'
        'This tests the "removeDuplicate" part of the algorithm',
        (WidgetTester tester) async {
          const _key = ValueKey('Stack');
          await tester.pumpWidget(
            MaterialApp(
              home: Scaffold(
                body: SizedBox(
                  width: 100,
                  height: 100,
                  child: Stack(
                    key: _key,
                    fit: StackFit.expand,
                    children: const [
                      // The two widget are placed at the bottom of the stack.
                      // The center of the stack will be hit, therefore the Align
                      // will be reported with the same TVP (the one ending at
                      // the Stack).
                      Align(
                        alignment: Alignment.bottomCenter,
                        child: Text('Text 1'),
                      ),
                      Align(
                        alignment: Alignment.bottomCenter,
                        child: Text('Text 2'),
                      ),
                    ],
                  ),
                ),
              ),
            ),
          );

          final hitTVPs = getTVPAtHitPosition(
            getCenterFromRenderBoxKey(key: _key),
          );
          final decompressedHitTVPs =
              hitTVPs.map((hitTVP) => decompressTVP(tvp: hitTVP)).toList();

          expect(
            decompressedHitTVPs.single,
            matchesFixture('hitTVPRemoveDuplicates'),
          );
        },
      );

      testWidgets(
        'Should return one TVP when a matched TVP are included in another when searching for widgets at the hit position\n'
        'This tests the "removeInclusion" part of the algorithm',
        (WidgetTester tester) async {
          const _key = ValueKey('Stack');
          await tester.pumpWidget(
            MaterialApp(
              home: Scaffold(
                body: SizedBox(
                  width: 100,
                  height: 100,
                  child: Stack(
                    key: _key,
                    fit: StackFit.expand,
                    children: const [
                      // The text won't be in the TVP because the middle of the Stack
                      // is hit. Therefore the TVP will end in the Stack, which is
                      // included in the TVP of the "SizedBox".
                      Align(
                        alignment: Alignment.bottomCenter,
                        child: Text('Text 1'),
                      ),

                      // The "SizedBox" widget will be hit and form a TVP including the
                      // one ending at the Stack.
                      Center(
                        child: SizedBox(height: 50, width: 50),
                      ),
                    ],
                  ),
                ),
              ),
            ),
          );

          final hitTVPs = getTVPAtHitPosition(
            getCenterFromRenderBoxKey(key: _key),
          );
          final decompressedHitTVPs =
              hitTVPs.map((hitTVP) => decompressTVP(tvp: hitTVP)).toList();

          expect(
            decompressedHitTVPs.single,
            matchesFixture('hitTVPRemoveInclusion'),
          );
        },
      );

      testWidgets(
        'Should return several TVPs when we cannot differentiate between them because the hit target returned by the hit test is a parent of both',
        (WidgetTester tester) async {
          const _key = ValueKey('Stack');
          await tester.pumpWidget(
            MaterialApp(
              home: Scaffold(
                body: SizedBox(
                  width: 100,
                  height: 100,
                  // The Stack will respond the the hitTest but the [SizedBox]s will
                  // not.
                  child: Stack(
                    key: _key,
                    fit: StackFit.expand,
                    children: const [
                      // This is in the tapped zone and therefore will be included
                      Center(
                        child: SizedBox(height: 50, width: 50),
                      ),

                      // This is in the tapped zone and therefore will be included
                      Center(
                        child: SizedBox(height: 50, width: 50),
                      ),
                    ],
                  ),
                ),
              ),
            ),
          );

          final hitTVPs = getTVPAtHitPosition(
            getCenterFromRenderBoxKey(key: _key),
          );
          final decompressedHitTVPs = hitTVPs
              .map(
                (hitTVP) => decompressTVP(tvp: hitTVP),
              )
              .toList();

          // The order does not matter, so we don't use "listEquals".
          expect(hitTVPs.length, 2);
          expect(
            decompressedHitTVPs[0],
            matchesFixture('multipleHitTVPs0'),
          );
          expect(
            decompressedHitTVPs[1],
            matchesFixture('multipleHitTVPs1'),
          );
        },
      );
    },
  );
}
