import 'dart:collection';
import 'dart:convert';

import 'package:contentsquare/src/modules/screengraph/painted_render_object_tree.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';
import 'package:flutter_test/flutter_test.dart';

import '../../../helpers/decompress_tvp.dart';
import '../../../helpers/matchers/fixtures/matcher.dart';
import 'fixtures/shared_fixture.dart';

/// Returns the [Type] of [T]
Type typeOf<T>() => T;

/// A widget which render object can be named using the generic parameter.
///
/// You can size the render object using the [size] parameter.
class WidgetWithNamedRenderObject<Name> extends SingleChildRenderObjectWidget {
  const WidgetWithNamedRenderObject({Key? key, Widget? child, this.size})
      : super(key: key, child: child);

  final Size? size;

  @override
  RenderObject createRenderObject(BuildContext context) {
    return _NamedRenderObject<Name>(size: size);
  }
}

class _NamedRenderObject<Name> extends RenderProxyBox {
  _NamedRenderObject({required Size? size}) : _size = size;

  final Size? _size;

  @override
  void performLayout() {
    final _size = this._size;
    if (_size == null) {
      return super.performLayout();
    }

    // Inspired by [RenderConstrainedBox]
    final child = this.child;
    if (child != null) {
      child.layout(
        BoxConstraints.tight(_size).enforce(constraints),
        parentUsesSize: true,
      );
    }
    size = _size;
  }
}

class _UnpaintedPage {}

class _PaintedPage {}

class _Untracked {}

class _Tracked1 {}

class _Tracked2 {}

mixin _PaintedRenderBox on RenderBox {
  Color get color;

  @override
  void paint(PaintingContext context, Offset offset) {
    // Inspired by [_RenderColoredBox]
    if (size > Size.zero) {
      context.canvas.drawRect(offset & size, Paint()..color = color);
    }
  }
}

class _RenderObjectWithElementlessRenderObjectChildWidget
    extends SingleChildRenderObjectWidget {
  const _RenderObjectWithElementlessRenderObjectChildWidget({
    Key? key,
    required this.size,
  }) : super(key: key);

  final Size size;

  @override
  RenderObject createRenderObject(BuildContext context) {
    return _RenderObjectWithElementlessRenderObjectChild(size: size);
  }
}

/// A render object which:
///   * Has a size of [size]
///   * Paints itself blue
///   * Creates a child render object that in not associated with any [Element]
///     which is red and has a size of half [size]
///
/// This class tries to mimic what [RenderEditable] does.
class _RenderObjectWithElementlessRenderObjectChild extends RenderConstrainedBox
    with _PaintedRenderBox {
  /// The child render object that is not associated with any [Element].
  late final backgroundRenderObject = _RedRenderBox(
    additionalConstraints: additionalConstraints / 2,
  );

  _RenderObjectWithElementlessRenderObjectChild({required Size size})
      : super(
          additionalConstraints: BoxConstraints.tightFor(
            width: size.width,
            height: size.height,
          ),
        ) {
    adoptChild(backgroundRenderObject);
  }

  @override
  Color get color => Colors.blueAccent;

  @override
  void markNeedsPaint() {
    super.markNeedsPaint();
    backgroundRenderObject.markNeedsPaint();
  }

  @override
  void attach(covariant PipelineOwner owner) {
    backgroundRenderObject.attach(owner);
    super.attach(owner);
  }

  @override
  void detach() {
    super.detach();
    backgroundRenderObject.detach();
  }

  @override
  void redepthChildren() {
    backgroundRenderObject.redepthChildren();
    super.redepthChildren();
  }

  @override
  void visitChildren(RenderObjectVisitor visitor) {
    visitor(backgroundRenderObject);
    super.visitChildren(visitor);
  }

  @override
  void dispose() {
    backgroundRenderObject.dispose();
    super.dispose();
  }

  @override
  void paint(PaintingContext context, Offset offset) {
    super.paint(context, offset);
    context.paintChild(backgroundRenderObject, offset);
  }

  @override
  void layout(Constraints constraints, {bool parentUsesSize = false}) {
    super.layout(constraints, parentUsesSize: parentUsesSize);
    backgroundRenderObject.layout(constraints, parentUsesSize: parentUsesSize);
  }
}

/// A render object which paints itself red.
class _RedRenderBox extends RenderConstrainedBox with _PaintedRenderBox {
  _RedRenderBox({required BoxConstraints additionalConstraints})
      : super(additionalConstraints: additionalConstraints);

  @override
  final Color color = Colors.redAccent;
}

main() {
  group(
    'getScreengraphFromRoot',
    () {
      testWidgets(
        'should return the correct screengraph after a scroll',
        (tester) async {
          await tester.pumpWidget(
            MaterialApp(
              home: Padding(
                padding: const EdgeInsets.all(10),
                child: Viewport(
                  offset: ViewportOffset.fixed(10),
                  slivers: [
                    SliverToBoxAdapter(
                      child: SizedBox(
                        height: 800,
                        child: Scaffold(
                          body: ListView.builder(
                            itemBuilder: (context, index) {
                              return ListTile(
                                title: SizedBox(
                                  height: 50,
                                  child: Text(
                                    'Text $index',
                                    key: ValueKey('Text $index'),
                                  ),
                                ),
                              );
                            },
                            itemCount: 30,
                          ),
                        ),
                      ),
                    ),
                  ],
                ),
              ),
            ),
          );

          // Scroll to invisible item
          final itemFinder = find.byKey(const ValueKey('Text 20'));
          await tester.scrollUntilVisible(itemFinder, 100);

          final screengraph = await getScreengraphFromRoot(
            nativePath: nativePath,
            pixelRatio: 1,
          );
          final decompressedScreengraph = decompressScreengraph(
            nativePath: nativePath,
            screengraphString: screengraph,
          );

          expect(
            decompressedScreengraph,
            matchesFixture('screengraphAfterScroll'),
          );
        },
      );

      testWidgets(
        'should take into account the pixel ratio',
        (tester) async {
          await tester.pumpWidget(
            const MaterialApp(
              home: Scaffold(
                body: Center(
                  child: Text('Hello World'),
                ),
              ),
            ),
          );

          final _screengraphPixelRatio05 = await getScreengraphFromRoot(
            nativePath: nativePath,
            pixelRatio: 0.5,
          );
          final decompressedScreengraphPixelRatio05 = decompressScreengraph(
            nativePath: nativePath,
            screengraphString: _screengraphPixelRatio05,
          );

          final _screengraphPixelRatio2 = await getScreengraphFromRoot(
            nativePath: nativePath,
            pixelRatio: 2,
          );
          final decompressedScreengraphPixelRatio2 = decompressScreengraph(
            nativePath: nativePath,
            screengraphString: _screengraphPixelRatio2,
          );

          expect(
            decompressedScreengraphPixelRatio05,
            matchesFixture('screengraphPixelRatio05'),
          );
          expect(
            decompressedScreengraphPixelRatio2,
            matchesFixture('screengraphPixelRatio2'),
          );
          expect(_screengraphPixelRatio05, isNot(_screengraphPixelRatio2));
        },
      );

      // It is possible that a painted render object has no associated element
      //
      // For example a [RenderEditable] creates child render object that are not
      // associated with any element. Therefore they will be present in the
      // painted tree but not in the element tree.
      //
      // In this case we don't add it in the screengraph since we can't associate
      // it with any path.
      testWidgets(
        'should not fail when a painted render object is not associated to any element',
        (tester) async {
          await tester.pumpWidget(
            const MaterialApp(
              home: Center(
                child: _RenderObjectWithElementlessRenderObjectChildWidget(
                  size: Size(200, 200),
                ),
              ),
            ),
          );

          // [getScreengraphFromRoot] should not throw an exception
          await expectLater(
            () => getScreengraphFromRoot(
              nativePath: nativePath,
              pixelRatio: 1,
            ),
            returnsNormally,
          );
        },
      );

      // Same render object can be present in the element tree but not in the
      // painted render object tree.
      //
      // In this case, they should not be included in the screengraph.
      testWidgets(
        'should not include RenderObjects which are not painted',
        (WidgetTester tester) async {
          await tester.pumpWidget(
            MaterialApp(
              home: Navigator(
                pages: const [
                  // The page below an opaque page is not painted
                  MaterialPage(
                    child: Center(
                      child: WidgetWithNamedRenderObject<_UnpaintedPage>(),
                    ),
                  ),

                  // The opaque page prevents the page below from being painted
                  MaterialPage(
                    child: Center(
                      child: WidgetWithNamedRenderObject<_PaintedPage>(),
                    ),
                  ),
                ],
                onPopPage: (route, result) => route.didPop(result),
              ),
            ),
          );

          final encodedScreengraphWithoutAllPaintedRenderObjects =
              await getScreengraphFromRoot(
            nativePath: nativePath,
            pixelRatio: 1,
          );
          final screengraphWithoutAllPaintedRenderObjects = _decodeScreengraph(
            encodedScreengraphWithoutAllPaintedRenderObjects,
          );

          expect(
            _screengraphContains<WidgetWithNamedRenderObject<_UnpaintedPage>>(
              screengraph: screengraphWithoutAllPaintedRenderObjects,
            ),
            isFalse,
          );

          expect(
            _screengraphContains<WidgetWithNamedRenderObject<_PaintedPage>>(
              screengraph: screengraphWithoutAllPaintedRenderObjects,
            ),
            isTrue,
          );
        },
      );

      testWidgets(
        'should not include render object which have the same frame as their parent in the screengraph',
        (WidgetTester tester) async {
          await tester.pumpWidget(
            const MaterialApp(
              home: Center(
                child: WidgetWithNamedRenderObject<_Tracked1>(
                  size: Size.square(200),
                  child: WidgetWithNamedRenderObject<_Untracked>(
                    size: Size.square(200),
                    child: Center(
                      child: WidgetWithNamedRenderObject<_Tracked2>(
                        size: Size.square(100),
                      ),
                    ),
                  ),
                ),
              ),
            ),
          );

          final encodedScreengraphWithoutSameFrameRenderObjects =
              await getScreengraphFromRoot(
            nativePath: nativePath,
            pixelRatio: 1,
          );

          final screengraphWithoutSameFrameRenderObjects = _decodeScreengraph(
            encodedScreengraphWithoutSameFrameRenderObjects,
          );

          expect(
            _screengraphContains<WidgetWithNamedRenderObject<_Tracked1>>(
              screengraph: screengraphWithoutSameFrameRenderObjects,
            ),
            isTrue,
          );

          expect(
            _screengraphContains<WidgetWithNamedRenderObject<_Tracked2>>(
              screengraph: screengraphWithoutSameFrameRenderObjects,
            ),
            isTrue,
          );

          expect(
            _screengraphContains<WidgetWithNamedRenderObject<_Untracked>>(
              screengraph: screengraphWithoutSameFrameRenderObjects,
            ),
            isFalse,
          );
        },
      );
    },
  );
}

/// Decodes [screengraphString]
UnmodifiableListView<Map<String, dynamic>> _decodeScreengraph(
    String screengraphString) {
  final decodedScreengraphList = jsonDecode(screengraphString) as List;
  return UnmodifiableListView(
      decodedScreengraphList.cast<Map<String, dynamic>>());
}

/// Utility class to visit nodes of [screengraph]
void _visitScreengraph({
  required UnmodifiableListView<Map<String, dynamic>> screengraph,
  required void Function(Map<String, dynamic>) onVisit,
}) {
  for (final node in screengraph) {
    onVisit(node);
    if (node['children'] != null) {
      _visitScreengraph(
        screengraph: UnmodifiableListView(
          (node['children'] as List).cast<Map<String, dynamic>>(),
        ),
        onVisit: onVisit,
      );
    }
  }
}

/// Utility class check that a [screengraph] contains a [Widget] of type [T] as
/// one of its nodes.
///
/// It does that by checking the 'class_name' property of the node.
bool _screengraphContains<T extends Widget>({
  required UnmodifiableListView<Map<String, dynamic>> screengraph,
}) {
  var found = false;
  _visitScreengraph(
    screengraph: screengraph,
    onVisit: (Map<String, dynamic> child) {
      if (child['metadata']?['class_name'] == typeOf<T>().toString()) {
        found = true;
      }
    },
  );
  return found;
}
